// 1.6.2007
//----------------------------------------------------------------------------------
//	Prevodnik RS232 <--> RS485
//	pri startbitu na RS232 zapne okamzite vysilac pokud nejsou detekovana data z RS485.
//----------------------------------------------------------------------------------
//Program je urcen k rizeni toku dat v prevodniku RS232 na RS485. S cilem eliminovat jakekoliv prvky nastavovene uzivatelem 
//bez nutnosti pouziti hardwaroveho rizeni toku. Program tedy dokaze detekovat prenosovou rychlost, paritni bit a stopbit.  
//
//
//Algoritmus:
//  Princip je zalozen na mereni nejkratsiho casoveho useku v osmi vzorcich, delka vzorku 
//  se odviji od velikosti prave nejmensiho zmereneho vzorku.
//  Vzorek se meri pomoci 8mi bitoveho casovace a interruptu od zmeny vstupu. casovac je osetren
//  proti pretekani (pocet preteceni se uklada, do promenne s ktere se zpetne vypocita 16bitove cislo, 
//  casovac se tedy virtualne chova jako 16ti bitovy.)
//  Mereni delky bitu 
//  
//  Aktualni implementace teto metody se chova tak, ze pokud je vyslan bajt ktery po startbitu obsahuje N 
//  stejnych bitu, tak je tato sekvence detekovana jako jeden bit to zpusobi ze detekovana rychlost je N krat 
//  mensi, nez skutecna. Dusledkem je, ze pokud se jeste pred stopbitem neobjevi samostatny bit s jinou hodnotou, 
//  nez predchozi sekvence. Tak bude vysln povoleno po dubu osminsobku zmen dlky "bitu" plus dal dva bity jako stopbity.
//  
//
//Vlastnosti:
//  Tento program je shopen detekovat pararitn bit, ale neni shopen detekovat poet stopbit.
//  To znamen e kdy bude parita logick jedna, tak bude rozpoznna jako stop bit.
//  Tato vlastnost ale pli nevad, protoe chovn algoritmu lze upravit tak, e bude predpokldat dva nebo ti stopbity.
//  V takovm ppad se packet penese korektn i kdy parita bude jedna, i nula.
//   
// Tento zpusob detekce neni imuni proti nahodnym chybovim spickam vzniklych v dusledku ruseni. 
// Proto je nutene napajeni kvalitne stbilizovat, pouzivat blokovaci kondenzatory a 
// zabezpecit, aby nedochazelo ke zvedani zeme.
////////////////////////////////////////////////////////////////////////////////////////
//
//Autor: Jakub Kakona (kaklik@mlab.cz)
//
////////////////////////////////////////////////////////////////////////////////////////
//TODO:
//  Optimalizovat kod, (hlavne najit casove vyhodnejsi umisteni pro nastavovani defaultnich hodnot promennych)
//  Bylo bydobre zavest uspavani pred prijetim bajtu.
//  Vykouset program na ATtiny13
//  Pri rychlostech mensich nez 9600kbps dochazi k pretekani promennych a algoritmus se nechova korektne.
//  Neni vyzkousen break, byly testovany pouze ASCII znaky vysilane pres terminal z klavesnice.  
//  Aktivovat WatchDog timer.
//  Neni vyzkouseno na vyssm taktu CPU jak 3,7MHz
//////////////////////////////////////////////////////////////////////////////////////
//


#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>

#define POVOLOVAK_LOW	PORTB &= ~(1<<PB3)   // nastaveni nuly na vystup, krery zapina povoleni vysilani RS485
#define POVOLOVAK_HIGH  PORTB |= (1<<PB3)    // nastaveni jednicky na tentyz vystup
#define DATA	(PIND & (1<<PD3))			// nastaveni vstupu, na kterem se budou detekovat data, musi byt stejny jako vstup pouziteho interruptu 

volatile unsigned int preteceni;  // promenna na ukladani poctu preteceni casovace
volatile unsigned int bit;		// promena pro pocitani bitu v bajtu
volatile unsigned long int bitdelay; // obsahuje aktualni zmereny cas delky bitu
volatile unsigned long int cas;		// urcuje nejkratsi nalezeny cas delky bitu

ISR(TIMER0_OVF_vect)   // interrupt od preteceni casovace
{
  preteceni++;			// kdyz pretece, poznamenej to.
}

ISR(INT1_vect)   // interrupt od zmeny vstupu na datech
{
  if ((bitdelay=(TCNT0+preteceni*0x0100)) < cas) cas = bitdelay; // provnani jestli zrovna zmereny cas je kratsi nez nejmensi znamy cas   
  TCNT0 = 0;					// zacni znova merit cas zacatku stopbitu
  preteceni=0;				// vynuluj vsechny casove promenne
  bit++;		// posun pocitadlo bitu o jednicku 
}

// ------------------------------------------------------------------
//	Main
// ------------------------------------------------------------------
int main(void)
{
int stopbit;
int ctvrtbit;

/*typedef struct {		// struktura pro celoseln dlen 
int quot;
int rem;
} div_t;*/


  DDRB |= (1<<DDB3);     // povoleni vystupu na povolovak 

  TIMSK |= (1 << TOIE0); // Enable timer overflow interrupt

  MCUCR |= (1 << ISC10); // nastaveni preruseni od zmeny vstupu pro interrupt INT0
  GIMSK |= (1 << INT1);  // povoleni interruptu INT0
   
  TCCR0B |= (1 << CS00);  // Set up timer 

  sei();   // enable interrupts

  while(1)
  {
	if(!DATA)   // kdyz je 0 na datech 
	{ 
	  POVOLOVAK_HIGH;    // zapni vysilani

      TCNT0=0;
      preteceni=0;
	  bit=0;			// vynuluj vsechny promenne pro mereni casu a pocitani bitu
	  cas = 0xFFFF;    // nastav cas delky bitu na nejvetsi mozny
	  	  	  	  
	  while (bit <= 8)  // odpocitej dobu 8mi bitu
	  {
	    if ((TCNT0+preteceni*0x0100) >= (cas + (ctvrtbit = cas >> 2))) // zkontroluj jestli necekame na dalsi bit dele nez je nekratsi znema delka bitu
// v kontrole je rezerva delky ctvrt bitu
		{
		  bit++;						// pokud jsme cekali dele, tak bit uz ubehl, to znamena, ze muzeme pocitadlo bitu posunout o jednicku
//		  div(ctvrtbit,0x0100);			// funkce pro implementaci celociselneho deleni zustava ale nepouzita kvuli nutnosti zavest datove struktury, bude ji nutno pouzit pro vetsi rychlosti   
          TCNT0 = ctvrtbit % 0x0100; //reminder; 					// zacni znova merit cas do dalsiho mozneho bitu 
          preteceni = ctvrtbit/0x0100; //quota;
		} 
	  }

	  while (!DATA);				// cekani na stop bit (detekce pripadneho paritniho bitu)

	  for (stopbit=2;stopbit >= 0;stopbit--)    // odpocitej dva stopbity
	  {
	    while ((TCNT0+preteceni*0x0100) <= cas) 
		{
		  if(!DATA) break;
        }
	  }
    }
    else  POVOLOVAK_LOW;  	// kdyz je 1 prepni na prijem 
  }
  return(0);
}
